/*-
 * Copyright (C) GD <gd@pikenet.ru>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: num2bit.c,v 1.3 2007/01/27 23:54:56 babolo Exp $
 */

#include <postgres.h>
#include <utils/varbit.h>
#include <funcapi.h>
#define fctx ((results*)funcctx->user_fctx)

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

typedef struct {
    int cur;
    VarBit* str;
} results;

PG_FUNCTION_INFO_V1(num2bit);

Datum 
num2bit(PG_FUNCTION_ARGS)
{
	VarBit         *newval, *oldval;
	uint32          port;
	int             rlen, newlen, ipad;
	bits8          *r, mask;

	oldval = PG_GETARG_VARBIT_P(0);

	if (PG_ARGISNULL(1))
		PG_RETURN_VARBIT_P(oldval);

	port = PG_GETARG_INT32(1);
	newlen = (port + 1 >= VARBITLEN(oldval)) ? port + 1 : VARBITLEN(oldval);

	rlen = VARBITTOTALLEN(newlen);
	newval = (VarBit *) palloc0(rlen);
	VARATT_SIZEP(newval) = rlen;
	VARBITLEN(newval) = newlen;

	memcpy(VARBITS(newval), VARBITS(oldval),
	       (VARBITBYTES(newval) <= VARBITBYTES(oldval)) ? VARBITBYTES(newval) : VARBITBYTES(oldval));

	r = VARBITS(newval);

	for (; port + 1 > BITS_PER_BYTE; r++, port -= BITS_PER_BYTE);
	*r |= (bits8) (1 << (BITS_PER_BYTE - (port + 1)));

	ipad = VARBITPAD(newval);
	if (ipad > 0) {
		mask = BITMASK << ipad;
		*(VARBITS(newval) + VARBITBYTES(newval) - 1) &= mask;
	}
	PG_RETURN_VARBIT_P(newval);
}


PG_FUNCTION_INFO_V1(bit2num);
Datum
bit2num(PG_FUNCTION_ARGS) {
    FuncCallContext *funcctx;
    MemoryContext    oldcontext;
    VarBit          *bitstr;
    int              i;

    if  (SRF_IS_FIRSTCALL()) {
        funcctx = SRF_FIRSTCALL_INIT();
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        if  (PG_ARGISNULL(0)) SRF_RETURN_DONE(funcctx);
        bitstr = PG_GETARG_VARBIT_P(0);
        fctx = (results*)palloc(sizeof(int) + VARBITTOTALLEN(VARBITLEN(bitstr)));
        if  (fctx == NULL)
            ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("bit2num: out of memory")));
        MemSet(fctx, 0, sizeof(int) + VARBITTOTALLEN(VARBITLEN(bitstr)));
        fctx->str = bitstr;
        fctx->cur = 0;
        MemoryContextSwitchTo(oldcontext);
    }
    funcctx = SRF_PERCALL_SETUP();
    for (i = fctx->cur; i < VARBITLEN(fctx->str); i++) {
        if  ( ((*(VARBITS(fctx->str) + i / BITS_PER_BYTE)) >> (BITS_PER_BYTE - (i % BITS_PER_BYTE) - 1)) & (bits8)1 ) {
            fctx->cur = i + 1;
            break;
    }   }
    if  ( fctx->cur == i + 1 ) SRF_RETURN_NEXT(funcctx, Int32GetDatum(i));
    else SRF_RETURN_DONE(funcctx);
}

